/*

    This file is part of the Maude 3 interpreter.

    Copyright 1997-2006 SRI International, Menlo Park, CA 94025, USA.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

*/

//
//	Forward declarations for strategy language classes.
//
#ifndef _strategyLanguage_hh_
#define _strategyLanguage_hh_

class StrategyExpression;
class TrivialStrategy;
class ApplicationStrategy;
class ConcatenationStrategy;
class IterationStrategy;
class BranchStrategy;
class UnionStrategy;
class TestStrategy;
class SubtermStrategy;
class CallStrategy;
class OneStrategy;
class ChoiceStrategy;
class SampleStrategy;

class StrategicExecution;

class StrategicProcess;
class DecompositionProcess;
class ApplicationProcess;
class MatchProcess;
class SubtermProcess;
class CallProcess;

class StrategicTask;
class StrategicSearch;
class BranchTask;
class RewriteTask;
class SubtermTask;
class CallTask;

class ModelCheckerSubtermTask;
class BiasedSubtermTask;
class FullSubtermTask;

class SharedRewriteSearchState;
class StrategyStackManager;
class VariableBindingsManager;

#endif
