/*
 * Autogenerated by the Meson build system.
 * Do not edit, your changes will be lost.
 */

#pragma once

/* Define to 1 if you have the <ctype.h> header file. */
#define HAVE_CTYPE_H 1

/* Define to 1 if you have the <limits.h> header file. */
#define HAVE_LIMITS_H 1

/* Define to 1 if you have the <math.h> header file. */
#define HAVE_MATH_H 1

/* Define to 1 if you have the <ostream> header file. */
#define HAVE_OSTREAM 1

/* Define to 1 if you have the <stddef.h> header file. */
#define HAVE_STDDEF_H 1

/* Define to 1 if you have the <stdio.h> header file. */
#define HAVE_STDIO_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the <sys/termios.h> header file. */
#define HAVE_SYS_TERMIOS_H 1

/* Define to 1 if you have the <sys/time.h> header file. */
#define HAVE_SYS_TIME_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* inline and discard selected local functions */
#define LOCAL_INLINES 1

/* do not do runtime checks */
#define NO_ASSERT 1

/* Name of package */
#define PACKAGE "Maude"

/* Define to the address where bug reports for this package should be sent */
#define PACKAGE_BUGREPORT "maude-bugs@lists.cs.illinois.edu"

/* Define to the full name and version of this package */
#define PACKAGE_STRING "Maude 3.5+se"

/* Define to the version of this package */
#define PACKAGE_VERSION "3.5+se"

/* The size of int, as computed by sizeof. */
#define SIZEOF_INT 4

/* The size of long, as computed by sizeof. */
#define SIZEOF_LONG 8

/* The size of short, as computed by sizeof. */
#define SIZEOF_SHORT 2

/* The size of unsigned long, as computed by sizeof. */
#define SIZEOF_UNSIGNED_LONG 8

/* The size of void *, as computed by sizeof. */
#define SIZEOF_VOID_P 8

/* use libsigsegv to handle segmentation faults */
#define USE_LIBSIGSEGV 1

/* use Python SMT solver library */
#define USE_PYSMT 1

/* Version number of package */
#define VERSION "3.5+se"

/* the probabilistic strategy language is included */
#define WITH_PROBABILISTIC_SLANG 1

