//
//	List of all descent functions supported by MetaLevelSmtOpSymbol.
//	Format
//		MACRO(symbols name, number of args)
//

MACRO(metaSmtCheck, 4)
MACRO(metaSmtSearch, 12)
MACRO(metaSmtSearchPath, 12)
